"""Methods and classes for iterating over (DICOM) files"""
from pathlib import Path

from pydicom.misc import is_dicom
from dicomindex.core import (
    read_dicom_file,
)
from dicomindex.exceptions import DICOMIndexError
from dicomindex.logs import get_module_logger

logger = get_module_logger("iterators")


class AllFiles:
    def __init__(self, path):
        """Walk through all files in path.

        Iterate per first-level folders to create a better
        progress indication (avoid returning thousands of second-level folders in
        succession, to be ignored by f.is_file(), halting progress for a while).
        """
        self.path = Path(path)

    def __iter__(self):
        logger.debug(f"Iterating over all files in {self.path}")
        for item in self.path.glob("*"):
            if item.is_file():
                yield item
            elif item.is_dir():
                logger.debug(f"Iterating over sub-path {item}")
                for x in item.rglob("*"):
                    if x.is_file():
                        yield x
                    else:
                        continue


class AllDICOMFiles:
    def __init__(self, path):
        """Path for each DICOM file in path recursively

        Notes
        -----
        Slow on slower filesystems as is performs an extra file open operation.
        Consider using AllDicomDatasets below
        """
        self.path = Path(path)

    def __iter__(self):
        return iter(
            file
            for file in (x for x in self.path.rglob("*") if x.is_file() and is_dicom(x))
        )


class AllDICOMDatasets:
    def __init__(self, path):
        """Dataset for each DICOM file in path recursively

        On slower filesystems this is faster than first getting AllDICOMFiles and
        then opening each file, because AllDICOMFiles does a (partial) read already.
        The reads are the expensive part.
        """
        self.path = Path(path)

    def __iter__(self):
        for x in (y for y in self.path.rglob("*") if y.is_file()):
            try:
                yield read_dicom_file(x)
            except DICOMIndexError:
                continue  # ignore error


class Folder:
    def __init__(self, base_path):
        self.base_path = Path(base_path)

    def folders(self):
        """All folders recursively"""
        return iter(x for x in self.base_path.rglob("*") if x.is_dir())



from typing import List, Optional

from pydicom import Dataset
from sqlalchemy import Float, ForeignKey, String
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from dicomindex.fields import InstanceLevel, SeriesLevel, StudyLevel
from dicomindex.types import (
    DICOMDate,
    DICOMDateTime,
    DICOMFlattenedString,
    DICOMIntegerString,
    DICOMMultipleString,
    DICOMName,
    DICOMSequence,
    DICOMTime,
    DICOMUID,
)


class Base(DeclarativeBase):
    pass


class Patient(Base):
    __tablename__ = "patient"
    PatientID: Mapped[str] = mapped_column(primary_key=True)
    studies: Mapped[List["Study"]] = relationship(
        back_populates="patient", cascade="all, delete-orphan"
    )

    @classmethod
    def init_from_dataset(cls, dataset: Dataset):
        """Try to fill all fields of this model with info from dataset"""
        return cls(PatientID=dataset.PatientID)


class Study(Base):
    __tablename__ = "study"
    StudyInstanceUID: Mapped[str] = mapped_column(primary_key=True)
    PatientID: Mapped[str] = mapped_column(ForeignKey("patient.PatientID"))
    patient: Mapped["Patient"] = relationship(back_populates="studies")
    series: Mapped[List["Series"]] = relationship(
        back_populates="study", cascade="all, delete-orphan"
    )

    # === DICOM fields ===
    ReferringPhysicianName: Mapped[Optional[str]] = mapped_column(DICOMName(192))
    PatientBirthTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    AccessionNumber: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    ReasonForPerformedProcedureCodeSequence: Mapped[Optional[str]] = mapped_column(
        DICOMSequence(265)
    )
    SourceApplicationEntityTitle: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(16)
    )
    IssuerOfPatientIDQualifiersSequence: Mapped[Optional[str]] = mapped_column(
        DICOMSequence(265)
    )
    InstanceAvailability: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(16)
    )
    PatientName: Mapped[Optional[str]] = mapped_column(DICOMName(192))
    InstitutionName: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    CurrentPatientLocation: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(64)
    )
    ProcedureCodeSequence: Mapped[Optional[str]] = mapped_column(DICOMSequence(265))
    IssuerOfPatientID: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    PatientSex: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    NumberOfStudyRelatedInstances: Mapped[Optional[str]] = mapped_column(
        DICOMIntegerString(12)
    )
    NumberOfStudyRelatedSeries: Mapped[Optional[str]] = mapped_column(
        DICOMIntegerString(12)
    )
    StudyTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    PatientBirthDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    ConfidentialityCode: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    InstitutionalDepartmentName: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(64)
    )
    StudyStatusID: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    ModalitiesInStudy: Mapped[Optional[str]] = mapped_column(DICOMMultipleString(16))
    StudyDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    SOPClassesInStudy: Mapped[Optional[str]] = mapped_column(DICOMUID(128))
    RequestedProcedureID: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(16)
    )
    IssuerOfAccessionNumberSequence: Mapped[Optional[str]] = mapped_column(
        DICOMSequence(265)
    )
    StudyDescription: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    StudyID: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    OtherPatientIDsSequence: Mapped[Optional[str]] = mapped_column(DICOMSequence(265))

    @classmethod
    def init_from_dataset(cls, dataset: Dataset):
        """Try to fill all fields of this model with info from dataset"""
        fields_to_transfer = StudyLevel.fields.union({"PatientID"})
        return cls(**{tag: dataset.get(tag) for tag in fields_to_transfer})


class Series(Base):
    __tablename__ = "series"
    SeriesInstanceUID: Mapped[str] = mapped_column(primary_key=True)
    StudyInstanceUID: Mapped[str] = mapped_column(ForeignKey("study.StudyInstanceUID"))
    study: Mapped["Study"] = relationship(back_populates="series")
    instances: Mapped[List["Instance"]] = relationship(
        back_populates="series", cascade="all, delete-orphan"
    )

    # === DICOM fields ===
    Laterality: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    AnatomicRegionSequence: Mapped[Optional[str]] = mapped_column(DICOMSequence(265))
    PerformedProcedureStepStartDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    PerformedProcedureStepDescription: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(64)
    )
    SeriesNumber: Mapped[Optional[str]] = mapped_column(DICOMIntegerString(12))
    Modality: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    PerformedProcedureStepStatus: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(16)
    )
    NumberOfSeriesRelatedInstances: Mapped[Optional[str]] = mapped_column(
        DICOMIntegerString(12)
    )
    BodyPartExamined: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    SeriesDescription: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    PerformedProcedureStepStartTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    ProtocolName: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    PresentationIntentType: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(16)
    )
    SeriesType: Mapped[Optional[str]] = mapped_column(DICOMMultipleString(16))
    SeriesTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    SmallestPixelValueInSeries: Mapped[Optional[float]] = mapped_column(Float(4))
    SpatialResolution: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(32))
    SeriesDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    PerformingPhysicianName: Mapped[Optional[str]] = mapped_column(DICOMName(192))
    OperatorsName: Mapped[Optional[str]] = mapped_column(DICOMName(192))
    FrameOfReferenceUID: Mapped[Optional[str]] = mapped_column(DICOMUID(128))

    @classmethod
    def init_from_dataset(cls, dataset: Dataset):
        """Try to fill all fields of this model with info from dataset"""
        fields_to_transfer = SeriesLevel.fields.union({"StudyInstanceUID"})
        return cls(**{tag: dataset.get(tag) for tag in fields_to_transfer})


class Instance(Base):
    __tablename__ = "instance"
    SOPInstanceUID: Mapped[str] = mapped_column(primary_key=True)
    SeriesInstanceUID: Mapped[str] = mapped_column(
        ForeignKey("series.SeriesInstanceUID")
    )
    series: Mapped["Series"] = relationship(back_populates="instances")
    path: Mapped[str] = mapped_column(String(512))

    # === DICOM fields ===
    ContentTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    StationName: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    DateOfLastCalibration: Mapped[Optional[str]] = mapped_column(DICOMDate())
    TransferSyntaxUID: Mapped[Optional[str]] = mapped_column(DICOMUID(128))
    Rows: Mapped[Optional[float]] = mapped_column(Float(4))
    Manufacturer: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    ContentLabel: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(16))
    ObservationDateTime: Mapped[Optional[str]] = mapped_column(DICOMDateTime())
    ContentDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    SOPClassUID: Mapped[Optional[str]] = mapped_column(DICOMUID(128))
    PresentationCreationDate: Mapped[Optional[str]] = mapped_column(DICOMDate())
    CorrectedImage: Mapped[Optional[str]] = mapped_column(DICOMMultipleString(16))
    InstitutionAddress: Mapped[Optional[str]] = mapped_column(String(1024))
    NumberOfFrames: Mapped[Optional[str]] = mapped_column(DICOMIntegerString(12))
    DeviceSerialNumber: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    TimeOfLastCalibration: Mapped[Optional[str]] = mapped_column(DICOMTime())
    ContentDescription: Mapped[Optional[str]] = mapped_column(DICOMFlattenedString(64))
    Columns: Mapped[Optional[float]] = mapped_column(Float(4))
    ContentCreatorName: Mapped[Optional[str]] = mapped_column(DICOMName(192))
    SoftwareVersions: Mapped[Optional[str]] = mapped_column(DICOMMultipleString(64))
    BitsAllocated: Mapped[Optional[float]] = mapped_column(Float(4))
    ManufacturerModelName: Mapped[Optional[str]] = mapped_column(
        DICOMFlattenedString(64)
    )
    InstanceNumber: Mapped[Optional[str]] = mapped_column(DICOMIntegerString(12))
    PresentationCreationTime: Mapped[Optional[str]] = mapped_column(DICOMTime())
    PixelPaddingValue: Mapped[Optional[float]] = mapped_column(Float(4))
    ConceptNameCodeSequence: Mapped[Optional[str]] = mapped_column(DICOMSequence(265))

    @classmethod
    def init_from_dataset(cls, dataset: Dataset, path: str):
        """Try to fill all fields of this model with info from dataset"""
        fields_to_transfer = InstanceLevel.fields.union({"SeriesInstanceUID"})
        param_dict = {tag: dataset.get(tag) for tag in fields_to_transfer}
        param_dict["path"] = path
        return cls(**param_dict)


class DICOMFileDuplicate(Base):
    """For storing files that share a SOPInstanceUID with another file"""

    __tablename__ = "duplicate_file"
    path: Mapped[str] = mapped_column(String(512), primary_key=True)
    SOPInstanceUID: Mapped[str] = mapped_column()


class NonDICOMFile(Base):
    """For skipping previously visited files"""

    __tablename__ = "non_dicom_file"
    path: Mapped[str] = mapped_column(String(512), primary_key=True)



"""DICOM fields that belong to Study, Series and Instance levels

Notes
-----
The classes below determine which DICOM tags get saved for which db objects.
Changing the fields below will require in the classes in dicomindex.orm
"""


class InstanceLevel:
    """attributes can be returned if query_level=INSTANCE"""

    fields = {
        "SOPInstanceUID",
        "SOPClassUID",
        "Rows",
        "Columns",
        "NumberOfFrames",
        "BitsAllocated",
        "ContentDate",
        "ContentTime",
        "ObservationDateTime",
        "ConceptNameCodeSequence",
        "ContentLabel",
        "ContentDescription",
        "PresentationCreationDate",
        "PresentationCreationTime",
        "ContentCreatorName",
        "TransferSyntaxUID",
        "InstanceNumber",
        "Manufacturer",
        "StationName",
        "ManufacturerModelName",
        "DeviceSerialNumber",
        "InstitutionAddress",
        "CorrectedImage",
        "SoftwareVersions",
        "DateOfLastCalibration",
        "TimeOfLastCalibration",
        "PixelPaddingValue",
    }


class SeriesLevel:
    """attributes can be returned if query_level=SERIES, or query_level=INSTANCE"""

    fields = {
        "SeriesInstanceUID",
        "Modality",
        "SeriesDate",
        "SeriesTime",
        "SeriesDescription",
        "Laterality",
        "AnatomicRegionSequence",
        "BodyPartExamined",
        "FrameOfReferenceUID",
        "PerformedProcedureStepDescription",
        "ProtocolName",
        "PerformingPhysicianName",
        "PerformedProcedureStepStartDate",
        "PerformedProcedureStepStartTime",
        "OperatorsName",
        "PerformedProcedureStepStatus",
        "PresentationIntentType",
        "SeriesNumber",
        "SeriesType",
        "SmallestPixelValueInSeries",
        "SpatialResolution",
        "NumberOfSeriesRelatedInstances",
    }


class SeriesLevelPromotable:
    """Instance level attributes that may be returned even if query_level=SERIES

    These attributes are promotable from the Image-level to the Series-level metadata
    if the attribute value for all SOP instances in that series contain the same
    value. Therefore, these attributes may be returned in a query_level=SERIES query,
    but only when this condition is satisfied.

    Notes
    -----
    Including any of these attributes as return keys in a Series-level query will
    cause the full study metadata to be loaded from disk, even if the corresponding
    attribute was promoted to the study summary MINT metadata. This means there is an
    extra read from disk in addition to the database, so the performance of such
    queries may be slightly slower than queries for other Series-level attributes.
    """

    fields = {
        "Manufacturer",
        "ManufacturerModelName",
        "CorrectedImage",
        "DeviceSerialNumber",
        "InstitutionAddress",
        "SoftwareVersions",
        "StationName",
        "DateOfLastCalibration",
        "TimeOfLastCalibration",
    }


class StudyLevel:
    """All fields that can be returned for query_level=STUDY, SERIES or INSTANCE.

    Notes
    -----
    From manual:
    Note that if a specified attribute is not in the study, the attribute will not
    be returned. If the specified attribute is in the study but does not have a
    value, it will be returned without a value.
    """

    fields = {
        "PatientID",
        "IssuerOfPatientID",
        "IssuerOfPatientIDQualifiersSequence",
        "PatientName",
        "PatientSex",
        "PatientBirthDate",
        "PatientBirthTime",
        "StudyInstanceUID",
        "AccessionNumber",
        "IssuerOfAccessionNumberSequence",
        "StudyID",
        "StudyDate",
        "StudyTime",
        "ModalitiesInStudy",
        "ReferringPhysicianName",
        "RequestedProcedureID",
        "StudyDescription",
        "InstanceAvailability",
        "InstitutionName",
        "StudyStatusID",
        "ConfidentialityCode",
        "ProcedureCodeSequence",
        "ReasonForPerformedProcedureCodeSequence",
        "OtherPatientIDsSequence",
        "InstitutionalDepartmentName",
        "NumberOfStudyRelatedSeries",
        "NumberOfStudyRelatedInstances",
        "InstanceAvailability",
        "SOPClassesInStudy",
        "CurrentPatientLocation",
        "SourceApplicationEntityTitle",
    }



"""Multithreaded stuff for optimized IO operations

Dicomindex needs to index DICOM after all. Best do it fast.
"""
from collections import namedtuple
from multiprocessing import Process, Queue
from queue import Empty

from dicomindex.logs import get_module_logger

logger = get_module_logger("threading")


EagerIteratorStatus = namedtuple("EagerIteratorStatus", ["visited", "has_finished"])
Sentinel = namedtuple("Sentinel", ["text"])


class EagerIterator:
    def __init__(self, iterator):
        """Iterator with a length which is updated by iterating in separate process
           length indicates the number of items in iterator

           Created of Path.rglob() which might take a long time to complete, does not
           take too much memory to deplete, and is quite useful to know the length of
           (how many files do I still need to process)


        Parameters
        ----------
        iterator:
            Iterator to iterate over in separate thread

        Notes
        -----
        Designed as closure:
        >>> for item in EagerIterator():
        >>>    print(item)

        """
        self.iterator = iterator
        self.generator = None
        self.process = None
        self.process_status: EagerIteratorStatus = EagerIteratorStatus(
            visited=0, has_finished=False
        )

        self.value_queue = Queue()
        self.message_queue = Queue()
        self.has_finished = False

    def _init_generator(self):
        self.process = Process(
            target=self.push_iter_to_queue,
            args=(self.iterator, self.value_queue, self.message_queue),
        )
        self.process.start()

        while self.more_values_to_expect():
            logger.debug(
                f"Queue size is {self.value_queue.qsize()} and "
                f"process finished is {self.process_status.has_finished} and "
                f"Message Queue size is {self.message_queue.qsize()} "
            )
            try:
                yield self.value_queue.get(timeout=0.1)  # timeout as fallback
            except Empty:
                continue

        logger.debug("Cleaning up. Joining iterator process")

        self.process.join(timeout=1)  # clean up

    def more_values_to_expect(self):
        """There will be more to get from value queue"""
        self.update_process_status()
        return self.value_queue.qsize() > 0 or not self.process_status.has_finished

    def update_process_status(self):
        if self.has_finished:
            return  # no updates needed, nothing will change
        if self.process_status.has_finished:
            self.has_finished = True

        while not self.message_queue.empty():
            try:
                self.process_status = self.message_queue.get(block=False)
            except Empty:
                continue  # no message, no problem.

    def __next__(self):
        if not self.generator:
            self.generator = self._init_generator()
        return next(self.generator)

    def __iter__(self):
        return self

    def __len__(self):
        self.update_process_status()
        return self.process_status.visited

    def __enter__(self):
        if not self.generator:
            self.generator = self._init_generator()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        logger.debug("Terminating path iterator process")
        if self.process.is_alive():
            logger.debug("Joining iterator process")
            self.process.join(timeout=60)

    @staticmethod
    def push_iter_to_queue(iterator, value_queue, message_queue):
        """Push all items in iter into value_queue, send item count through
        message_queue

        If run in a separate process, can determine the size of a slow iterator as
        quickly as possible without having to wait for depletion.
        """
        visited = 0
        for item in iterator:
            visited += 1
            value_queue.put(item)
            message_queue.put(
                EagerIteratorStatus(visited=visited, has_finished=False), timeout=60
            )
        message_queue.put(
            EagerIteratorStatus(visited=visited, has_finished=True), timeout=60
        )
        logger.debug("push_iter_to_queue finished")



"""Additions to the pydicom Dataset object"""
from typing import Any, Optional, Tuple, Union

from pydicom.dataset import Dataset
from pydicom.tag import BaseTag

from dicomindex.exceptions import DICOMIndexError


class RequiredDataset(Dataset):
    """A pydicom Dataset,that raises distinctive errors when accessing missing keys

    Made this to specifically handle missing keys on a dataset. By default, a Dataset
    instance raises KeyError and AttributeError. These are too general to safely
    catch over larger pieces of code. Putting try-except blocks around each individual
    dict key access is ugly and annoying.

    Raises
    ------
    RequiredTagNotFound
        When a requested key is not found in this dataset. Either through attribute
        access, like dataset.PatientID or through dict access like
        dataset['PatientID']

    Notes
    -----
    Init like this:

    >>> ds = Dataset()
    >>> rds = RequiredDataset(ds)

    Now you can handle missing keys cleanly without accidentally catching other
    KeyErrors:

    >>> try:
    >>>     important_dataset_check(rds)
    >>> except RequiredTagNotFound:
    >>>     print('check failed due to missing information')

    """

    def __getattr__(self, name):
        try:
            return super().__getattr__(name)
        except AttributeError as e:
            raise RequiredTagNotFound(f"Required tag not found: {e}") from e

    def __getitem__(self, key):
        try:
            return super().__getitem__(key)
        except KeyError as e:
            raise RequiredTagNotFound(f"Required tag not found: {e}") from e

    def get(
        self,
        key: Union[str, Union[int, Tuple[int, int], BaseTag]],
        default: Optional[Any] = None,
    ):
        """Pass to Dataset but handle default value here.

        Needed because Dataset.get() internally relies on __getattr__ and __getitem__
        to return their standard exceptions.
        """
        try:
            return super().get(key, default)
        except RequiredTagNotFound:
            return default


class RequiredTagNotFound(DICOMIndexError):
    pass



"""Custom SQLAlchemy data types

Translates between pydicom datatypes and sqlalchemy fields.
Mainly just 'flatten complex pydicom data type to string' currently
"""

from datetime import datetime
import sqlalchemy.types as types


class DICOMSequence(types.TypeDecorator):
    """Saves pydicom Sequence objects"""

    impl = types.String

    cache_ok = True

    def process_bind_param(self, value, dialect):
        # just flatten and truncate I just want this in db, nothing fancy for now
        if value is None:
            return ""
        else:
            return str(value)[:255]

    def process_result_value(self, value, dialect):
        return value


class DICOMDate(types.TypeDecorator):
    """Dicom date string. Like 20120425 (VR = DA)"""

    impl = types.Date

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value:
            return datetime.strptime(value, "%Y%m%d")


class DICOMDateTime(types.TypeDecorator):
    """Dicom date string. Like 2012042514350204.123 (VR = DT)"""

    impl = types.DateTime

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value:
            return datetime.strptime(value, "%Y%m%d%H%M%S")


class DICOMTime(types.TypeDecorator):
    """Dicom date string. Like 14350204.123 (VR = TM)"""

    impl = types.Time

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value:
            if "." not in value:  # make strptime below work even without ms
                value = value + ".0"
            return datetime.strptime(value, "%H%M%S.%f").time()


class DICOMFlattenedString(types.TypeDecorator):
    """Converts complex DICOM value to string"""

    impl = types.String

    cache_ok = True

    def process_bind_param(self, value, dialect):
        # just flatten. I just want this in db, nothing fancy for now
        if value is None:
            return ""
        else:
            return str(value)

    def process_result_value(self, value, dialect):
        return value


class DICOMName(DICOMFlattenedString):
    """For DICOM PersonName"""


class DICOMUID(DICOMFlattenedString):
    """For DICOM UniqueIdentifier"""


class DICOMIntegerString(DICOMFlattenedString):
    """For DICOM IntegerString"""


class DICOMMultipleString(DICOMFlattenedString):
    """For DICOM objects with VM (Value Multiplicity) > 1"""



from collections import namedtuple
from typing import List


class PathStatuses:
    PROCESSED = "PROCESSED"
    SKIPPED_ALREADY_VISITED = "SKIPPED_ALREADY_VISITED"
    SKIPPED_FAILED = "SKIPPED_FAILED"
    SKIPPED_NON_DICOM = "SKIPPED_NON_DICOM"
    VISITED = "VISITED"


PathStatus = namedtuple("PathStatus", ["path", "status"])


class Statistics:
    """Records statistics for a typical dicomindex run"""

    def __init__(self):
        self.status_list: List[PathStatus] = []

    def add(self, path, status):
        self.status_list.append(PathStatus(path=path, status=status))

    def skipped_failed(self):
        return [x for x in self.status_list if x.status == PathStatuses.SKIPPED_FAILED]

    def skipped_non_dicom(self):
        return [
            x for x in self.status_list if x.status == PathStatuses.SKIPPED_NON_DICOM
        ]

    def processed(self):
        return [x for x in self.status_list if x.status == PathStatuses.PROCESSED]

    def skipped(self):
        return [
            x
            for x in self.status_list
            if x.status
            in (
                PathStatuses.SKIPPED_ALREADY_VISITED,
                PathStatuses.SKIPPED_FAILED,
                PathStatuses.SKIPPED_NON_DICOM,
            )
        ]

    def skipped_visited(self):
        return [
            x
            for x in self.status_list
            if x.status == PathStatuses.SKIPPED_ALREADY_VISITED
        ]

    def visited(self):
        return self.status_list

    def summary(self):
        return (
            f"Visited {len(self.visited())}, processed {len(self.processed())},"
            f" skipped {len(self.skipped())} "
            f"(already-visited/non-dicom/failed: {len(self.skipped_visited())}/"
            f"{len(self.skipped_non_dicom())}/"
            f"{len(self.skipped_failed())})"
        )



from sqlalchemy import create_engine
from sqlalchemy.orm import Session

from dicomindex.orm import Base


class SQLiteSession:
    """A database session on a sqlite file

    Examples
    --------
    with SQLiteSession('my_file.sql') as session:
        session.do_things()

    # closes after leaving closure
    """

    def __init__(self, db_path):
        self.db_path = db_path
        self.session = None

    def __enter__(self):
        engine = create_engine(f"sqlite:///{self.db_path}", echo=False)
        Base.metadata.create_all(engine, checkfirst=True)  # Create if needed
        self.session = Session(engine)
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.session:
            self.session.close()



import pydicom
from pydicom.errors import InvalidDicomError

from dicomindex.exceptions import NotDICOMError


def read_dicom_file(path):
    """

    Parameters
    ----------
    path: Path
        Read this path as dicom file

    Raises
    ------
    DICOMIndexError
        When given path cannot be opened as DICOM
    FileNotFoundError
        When given path does not exist

    """
    try:
        return pydicom.filereader.dcmread(str(path), stop_before_pixels=True)
    except InvalidDicomError as e:
        raise NotDICOMError(e) from e



import logging
from pathlib import Path

import click
from click import Path as ClickPath
from tabulate import tabulate  # type: ignore
from tqdm import tqdm

from dicomindex.persistence import SQLiteSession
from dicomindex.processing import index_folder_full, index_one_file_per_folder

from dicomindex.logs import get_module_logger
from dicomindex.orm import (
    DICOMFileDuplicate,
    Instance,
    NonDICOMFile,
    Patient,
    Series,
    Study,
)

logger = get_module_logger("cli")


def configure_logging(verbose):
    if verbose == 0:
        logging.basicConfig(level=logging.INFO)
        logging.info("Set loglevel to INFO")
    if verbose >= 1:
        logging.basicConfig(level=logging.DEBUG)
        logging.info("Set loglevel to DEBUG")


@click.group()
@click.option("-v", "--verbose", count=True)
def main(verbose):
    r"""DICOM index - see what dicom is in your folders

    Use the commands below with -h for more info
    """
    configure_logging(verbose)


@click.command(name="index_full")
@click.argument("index_file", type=ClickPath())
@click.argument("base_folder", type=ClickPath(exists=True))
def index_full_func(index_file, base_folder):
    """Index all files in base folder. Slow but complete"""
    base_folder = Path(base_folder)

    logger.info(f"Starting index of '{base_folder}'. Writing to '{index_file}'")
    if Path(index_file).exists():
        click.confirm(
            f'file "{index_file}" already exists. Do you want to '
            f"append to this file?",
            abort=True,
        )

    with SQLiteSession(index_file) as session:
        with tqdm(total=1) as pbar:
            stats = index_folder_full(base_folder, session, progress_bar=pbar)

    logger.info("Finished")
    logger.info(stats.summary())


@click.command(name="index_per_folder")
@click.argument("index_file", type=ClickPath())
@click.argument("base_folder", type=ClickPath(exists=True))
def index_per_folder_func(index_file, base_folder):
    """Index one file in each folder. Fast indexing but assumes files for the same
    patient/study/series are in one folder each. Might miss a lot of data is not
    of that structure
    """
    base_folder = Path(base_folder)

    logger.info(f"Starting index of '{base_folder}'. Writing to '{index_file}'")
    if Path(index_file).exists():
        click.confirm(
            f'file "{index_file}" already exists. Do you want to '
            f"append to this file?",
            abort=True,
        )

    with SQLiteSession(index_file) as session:
        with tqdm(total=1) as pbar:
            stats = index_one_file_per_folder(base_folder, session, progress_bar=pbar)

    logger.info("Finished")
    logger.info(stats.summary())


@click.command(name="stats")
@click.argument("index_file", type=ClickPath(exists=True))
def stats_func(index_file):
    """What is in the given index file?"""

    logger.debug(f"Loading '{index_file}'")
    with SQLiteSession(index_file) as session:
        print(f"Statistics for {index_file}")
        table = {
            "Patients": [session.query(Patient).count()],
            "Studies": [session.query(Study).count()],
            "Series": [session.query(Series).count()],
            "Instances:": [session.query(Instance).count()],
            "Duplicates:": [session.query(DICOMFileDuplicate).count()],
            "Non_dicom:": [session.query(NonDICOMFile).count()],
        }

        print(tabulate(table, headers="keys", tablefmt="simple"))


main.add_command(index_full_func)
main.add_command(index_per_folder_func)
main.add_command(stats_func)



"""Functions and classes for indexing dicom files in a folder

Ties together modules
"""
from typing import Dict, Optional, Set

from pydicom import Dataset
from sqlalchemy.exc import StatementError
from tqdm import tqdm

from dicomindex.core import read_dicom_file
from dicomindex.dataset import RequiredDataset, RequiredTagNotFound
from dicomindex.exceptions import NotDICOMError
from dicomindex.iterators import AllFiles, Folder
from dicomindex.logs import get_module_logger
from dicomindex.orm import (
    DICOMFileDuplicate,
    Instance,
    NonDICOMFile,
    Patient,
    Series,
    Study,
)
from dicomindex.statistics import PathStatuses, Statistics
from dicomindex.threading import EagerIterator

logger = get_module_logger("processing")


def process_path(path, session, index):
    """Add this path to db and take care of all bookkeeping

    Returns
    -------
    str
        One of the values in PathStatuses
    """

    path = str(path)
    if path in index.paths:
        return PathStatuses.SKIPPED_ALREADY_VISITED
    try:
        ds = read_dicom_file(path)
    except NotDICOMError:
        index.set_path_status(path, status=PathStatuses.SKIPPED_NON_DICOM)
        session.add(NonDICOMFile(path=path))
        session.commit()
        return PathStatuses.SKIPPED_NON_DICOM

    try:
        to_add = index.create_new_db_objects(
            RequiredDataset(ds), path, add_to_index=False
        )
    except RequiredTagNotFound as e:
        # a basic tag like PatientID is missing from this file. Don't process.
        logger.exception(e)
        logger.error(f"DICOM file at {path} was missing essential tags. skipping")
        index.set_path_status(path, PathStatuses.SKIPPED_FAILED)
        return PathStatuses.SKIPPED_FAILED

    session.add_all(to_add)
    try:
        session.commit()
    except StatementError as e:
        session.rollback()
        # Skip error, continue
        logger.exception(e)
        logger.error(f"Error processing {path}. skipping")
        return PathStatuses.SKIPPED_FAILED

    index.add_to_index(ds)  # commit has succeeded. Now you can add
    index.set_path_status(path, PathStatuses.PROCESSED)
    return PathStatuses.PROCESSED


def update_progress(
    iter_in: EagerIterator, stats: Statistics, progress_bar: Optional[tqdm]
):
    """Updates both count and total length of given tqdm progress bar"""
    if progress_bar:
        progress_bar.total = len(iter_in)
        progress_bar.update()
        progress_bar.set_postfix_str(stats.summary())


def index_folder_full(base_folder, session, progress_bar: Optional[tqdm] = None):
    """Iterate over all files in base folder, add them to session"""
    index = DICOMIndex.init_from_session(session)
    statistics = Statistics()
    logger.debug(f"Found {len(index.paths)} instances already in index")

    with EagerIterator(AllFiles(base_folder)) as paths:
        for path in paths:
            statistics.add(path, process_path(path, session, index))
            update_progress(paths, statistics, progress_bar)

    return statistics


def index_one_file_per_folder(
    base_folder, session, progress_bar: Optional[tqdm] = None
):
    """Add one file from each folder to index. This assumes the structure of base_folder
    somehow corresponds to patient/study/series.

    Notes
    -----
    This method will keep visiting files in a folder until a single DICOM file has
    been registered. Any non-dicom files or erroring files encountered during this
    search are registered in DB. This means a variable number of files per folder
    will be in db.
    """
    index = DICOMIndex.init_from_session(session)
    statistics = Statistics()
    logger.debug(f"Found {len(index.paths)} instances already in index")

    with EagerIterator(Folder(base_folder).folders()) as folders:
        for folder in folders:
            logger.debug(f"Finding single file in {folder}")
            folder_processed = False
            folder_skipped = False
            for path in (x for x in folder.glob("./*") if x.is_file()):
                if index.paths.get(str(path)) == PathStatuses.PROCESSED:
                    logger.debug(
                        f"Path {path} already processed. Skipping this " f"folder"
                    )
                    folder_skipped = True
                    folder_processed = True
                    break
                status = process_path(path, session, index)
                if status == PathStatuses.PROCESSED:
                    # added one., We're done with this folder
                    statistics.add(path, status)
                    update_progress(folders, statistics, progress_bar)
                    folder_processed = True
                    break
                else:
                    continue
            if not folder_processed:
                logger.debug(f"Found no DICOM files in {folder}")
                statistics.add(folder, PathStatuses.SKIPPED_NON_DICOM)
                update_progress(folders, statistics, progress_bar)
            elif folder_skipped:
                statistics.add(path, PathStatuses.SKIPPED_NON_DICOM)
                update_progress(folders, statistics, progress_bar)

    return statistics


class DICOMIndex:
    """Keeps track DICOM object ids to minimize database traffic

    A single DICOM Patient can be referenced in hundreds of DICOM instances. A
    Patient object will only have to be added to database once. Instead of
    checking the database hundreds of times for patients, keep track here.
    """

    def __init__(
        self,
        patient_ids: Set[str],
        study_uids: Set[str],
        series_uids: Set[str],
        instance_uids: Set[str],
        paths: Dict[str, str],
    ):
        self.patient_ids = patient_ids
        self.study_uids = study_uids
        self.series_uids = series_uids
        self.instance_uids = instance_uids
        self.paths = paths

    @classmethod
    def init_from_session(cls, session):
        """Populate indexes by reading ids from given session"""

        return cls(
            patient_ids={pid for pid, in session.query(Patient.PatientID)},
            study_uids={pid for pid, in session.query(Study.StudyInstanceUID)},
            series_uids={pid for pid, in session.query(Series.SeriesInstanceUID)},
            instance_uids={pid for pid, in session.query(Instance.SOPInstanceUID)},
            paths=dict(
                **{
                    str(path): PathStatuses.PROCESSED
                    for path, in session.query(Instance.path)
                },
                **{
                    str(path): PathStatuses.SKIPPED_FAILED
                    for path, in session.query(DICOMFileDuplicate.path)
                },
                **{
                    str(path): PathStatuses.SKIPPED_NON_DICOM
                    for path, in session.query(NonDICOMFile.path)
                },
            ),
        )

    def create_new_db_objects(self, dataset: Dataset, path: str, add_to_index=True):
        """Create patient/study/series/instance objects from dataset, ignore existing

        Notes
        -----
        Any file with previously seen SOPInstanceUID will be stored as duplicate
        and further processing will be skipped. This means any patient/study/series
        information in such a file is not recorded. Having duplicate SOPInstanceUIDs
        in an archive should never happen, but does.

        Parameters
        ----------
        dataset: Dataset
            Copy relevant fields from this dataset to db objects
        path: str
            Set Instance.path to this
        add_to_index: Bool, optional
            If True, automatically add ids and paths of any db objects to this index
            Defaults to true.


        Returns
        -------
        List[Base]
            Patient/Study/Series/Instance objects for ids that have not been

        """
        path = str(path)  # handle Path type input

        if dataset.SOPInstanceUID in self.instance_uids:
            logger.debug(
                f"SOPInstanceUID in '{path}' already exists. This is a"
                f" duplicate file. Storing in duplicate instances"
            )
            self.set_path_status(path)
            return [
                DICOMFileDuplicate(path=path, SOPInstanceUID=dataset.SOPInstanceUID)
            ]

        db_objects = [Instance.init_from_dataset(dataset, path)]

        if dataset.PatientID not in self.patient_ids:
            db_objects.append(Patient.init_from_dataset(dataset))
        if dataset.StudyInstanceUID not in self.study_uids:
            db_objects.append(Study.init_from_dataset(dataset))
        if dataset.SeriesInstanceUID not in self.series_uids:
            db_objects.append(Series.init_from_dataset(dataset))

        if add_to_index:
            self.add_to_index(dataset)
            self.set_path_status(path)
        logger.debug(f"Created {len(db_objects)} db objects for {path}")
        return db_objects

    def set_path_status(self, path, status: str = PathStatuses.VISITED):
        """Register the given path as visited"""
        self.paths[str(path)] = status

    def add_to_index(self, dataset):
        """Add patient/study/series/instance ids to index"""
        self.patient_ids.add(dataset.PatientID)
        self.study_uids.add(dataset.StudyInstanceUID)
        self.series_uids.add(dataset.SeriesInstanceUID)
        self.instance_uids.add(dataset.SOPInstanceUID)
